import random
#import matplotlib.pyplot as plt
import numpy as np
import time
from torchvision.datasets import ImageFolder
import sys
sys.path.append('..')
# from utils import *
from torch.autograd import Variable
import numpy as np
from tqdm import tqdm
import os
import torch
import torch.nn as nn
import random
import time
from torchvision import transforms
from torch.utils.data import DataLoader
from torchvision import models
import torch.optim as optim
from torch.optim import lr_scheduler
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score
import gc
import warnings
temp = sys.stdout
#In this program, you need to change the "data_index = 0" and "x = [48,  12]" sections, "data_index" indicates the index of the target dataset,
# "x = [12,  48]" represents the number of layers fine-tuned in the two source domain models and is obtained by Procedure "AMLS.py"
#the output file is in folder "outfile"

# 把输出重定向到文件
f = open('outfile/code1.log', 'w')

# 之后使用print函数，都将内容打印到 screenshot.log 文件中
sys.stdout = f
warnings.filterwarnings("ignore")

def setup_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

def train_one_batch_finetune(images, labels, model1, model2, criterion):
    '''
    运行一个batch的训练，返回当前batch的训练日志
    :param images:
    :param labels:
    :return:
    '''
    #model = torch.nn.DataParallel(model)
    images = images.to(device)
    labels = labels.to(device)
    outputs1 = model1(images)
    outputs2 = model2(images)
    ###############
    labels_onehot = torch.zeros_like(outputs1)  # 创建与outputs形状相同的one-hot编码张量
    labels_onehot.scatter_(1, labels.unsqueeze(1), 1)  # 对one-hot编码的正确类别位置赋值为1
    loss = criterion(outputs1, outputs2, labels_onehot.float()) # 计算自定义MSE损失函数
    ####################
    # loss = criterion(outputs, labels)
    optimizer1.zero_grad()
    optimizer2.zero_grad()
    loss.backward()
    optimizer1.step()
    optimizer2.step()
    _,preds = torch.max(outputs1 + outputs2, 1)
    preds = preds.cpu().numpy()
    loss = loss.detach().cpu().numpy()
    # outputs = outputs.detach().cpu().numpy()
    labels = labels.detach().cpu().numpy()
    log_train = {}
    # log_train['epoch'] = epoch
    # log_train['batch'] = batch_idx
    log_train['train_loss'] = loss
    # log_train['train_accuracy'] = accuracy_score(labels, preds)
    log_train['preds'] = preds
    log_train['labels'] = labels

    return log_train
def evaluate_testset(model1, model2, criterion):
    loss_list = []
    labels_list = []
    preds_list = []

    with torch.no_grad():
        for images, labels in test_loader:
            images = images.to(device)
            labels = labels.to(device)
            outputs1 = model1(images)
            outputs2 = model2(images)

            _,preds = torch.max(outputs1 + outputs2,1)
            preds = preds.cpu().numpy()
            #########################
            labels_onehot = torch.zeros_like(outputs1)  # 创建与outputs形状相同的one-hot编码张量
            labels_onehot.scatter_(1, labels.unsqueeze(1), 1)  # 对one-hot编码的正确类别位置赋值为1
            loss = criterion(outputs1 , outputs2, labels_onehot.float())  # 计算自定义MSE损失函数
            ######################
            # loss = criterion(outputs, labels)
            loss = loss.detach().cpu().numpy()
            # outputs = outputs.detach().cpu().numpy()
            labels = labels.detach().cpu().numpy()

            loss_list.append(loss)
            labels_list.extend(labels)
            preds_list.extend(preds)
    log_test = {}
    # log_test['epoch'] = epoch

    log_test['test_loss'] = np.mean(loss_list)
    log_test['test_accuracy'] = accuracy_score(labels_list, preds_list)
    log_test['test_precision'] = precision_score(labels_list, preds_list, average = 'macro')
    log_test['test_recall'] = recall_score(labels_list, preds_list, average='macro')
    log_test['test_precision'] = f1_score(labels_list, preds_list, average='macro')

    return log_test
#the Multi-source Collaborative Loss Function
class My_loss(nn.Module):
    def __init__(self):
        super(My_loss, self).__init__()

    def forward(self, outputs1, outputs2, labels):
        # 对输入进行softmax操作
        outputs_softmax1 = torch.softmax(outputs1, dim=1)
        outputs_softmax2 = torch.softmax(outputs2, dim=1)
        # 计算交叉熵损失
        DE_loss1 = -torch.sum(labels * torch.log(outputs_softmax1 + 1e-8)) / labels.size(0)
        DE_loss2 = -torch.sum(labels * torch.log(outputs_softmax2 + 1e-8)) / labels.size(0)
        Dif_loss = 0.2 * torch.mean(torch.norm(outputs1-outputs2, p=2, dim=1))
        Total_loss = DE_loss1 + DE_loss2 + Dif_loss
        # dis_loss = torch.mean(torch.norm(outputs_softmax1-outputs_softmax2, p=2, dim=1))
        # loss = loss1 + loss2 + dis_loss
        return Total_loss


data_name = ['indoor', 'stf_dog', 'aircraft', 'ucf101', 'omniglot40','caltech256-30','caltech256-60']
num_classes = [67, 120, 100, 101, 1623,256,256]
data_index = 0
print(data_name[data_index])

seed = 0
setup_seed(seed)
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
batch_size = 64
data_dir = "./datasets/" + data_name[data_index] + '/'
normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
train_augs = transforms.Compose([
    transforms.Resize(256),
    transforms.RandomResizedCrop(224),
    transforms.RandomHorizontalFlip(),
    transforms.ToTensor(),
    normalize
])
test_augs = transforms.Compose([
    transforms.Resize(256),
    transforms.CenterCrop(size=224),
    transforms.ToTensor(),
    normalize
])
train_loader = DataLoader(ImageFolder(os.path.join(data_dir, 'train'), transform=train_augs),
                          batch_size, shuffle=True)
test_loader = DataLoader(ImageFolder(os.path.join(data_dir, 'val'), transform=test_augs),
                         batch_size, shuffle=False, )

x = [12,  48]
setup_seed(seed)
pretrained_net1 = models.resnet50(pretrained = False)
pathfile = 'pretrained_models/pre_model_imagenet_resnet50.pth'
pretrained_net1.load_state_dict(torch.load(pathfile))
in_features = pretrained_net1.fc.in_features
pretrained_net1.fc = nn.Linear(in_features, num_classes[data_index])
# pretrained_net.fc = nn.Linear(2048,num_classes[data_index])
pretrained_net1 = pretrained_net1.to(device)
lr, num_epochs = 0.01, 110

layer_num = -1
layer_tune = int(x[0])
for name, params in pretrained_net1.named_parameters():
    if 'conv' in name:
        layer_num += 1
    if layer_num<layer_tune or layer_num==layer_tune:
        break
    else:
        params.requires_grad = False

global optimizer1
optimizer1 = optim.SGD(filter(lambda p: p.requires_grad, pretrained_net1.parameters()), momentum=0.9, lr=lr, weight_decay=0.0005)
lr_scheduler1 = lr_scheduler.StepLR(optimizer1, step_size=30, gamma=0.1)

# pretrained_net2 = models.resnet50(pretrained=True)
pretrained_net2 = models.resnet50(pretrained = False)
pathfile = 'pretrained_models/pre_model_cifar100_resnet50.pth'
pretrained_net2 = torch.load(pathfile)
in_features = pretrained_net2.fc.in_features
pretrained_net2.fc = nn.Linear(in_features, num_classes[data_index])
# pretrained_net.fc = nn.Linear(2048,num_classes[data_index])
pretrained_net2 = pretrained_net2.to(device)

layer_num = -1
layer_tune = int(x[1])
for name, params in pretrained_net2.named_parameters():
    if 'conv' in name:
        layer_num += 1
    if layer_num<layer_tune or layer_num==layer_tune:
        break
    else:
        params.requires_grad = False

global optimizer2
optimizer2 = optim.SGD(filter(lambda p: p.requires_grad, pretrained_net2.parameters()), momentum=0.9, lr=lr, weight_decay=0.0005)
criterion = My_loss()#nn.CrossEntropyLoss()
lr_scheduler2 = lr_scheduler.StepLR(optimizer2, step_size=30, gamma=0.1)


test_accuracy_list = []
for epoch in range(1, num_epochs + 1):
    start = time.time()
    print(f'Epoch {epoch}/{num_epochs}')
    pretrained_net1.train()
    pretrained_net2.train()
    train_loss_list = []
    train_pre_list = []
    train_label_list = []
    for images, labels in tqdm(train_loader):
        log_train = train_one_batch_finetune(images, labels, pretrained_net1, pretrained_net2, criterion)
        train_loss_list.append(log_train['train_loss'])
        train_pre_list.extend(log_train['preds'])
        train_label_list.extend(log_train['labels'])
    train_loss = np.mean(train_loss_list)
    train_accuracy = accuracy_score(train_label_list, train_pre_list)
    lr_scheduler1.step()
    lr_scheduler2.step()
    pretrained_net1.eval()
    pretrained_net2.eval()
    log_test = evaluate_testset(pretrained_net1, pretrained_net2, criterion)
    test_accuracy = log_test['test_accuracy']
    test_accuracy_list.append(test_accuracy)
    test_loss = log_test['test_loss']
    print(f'Epoch {epoch}/{num_epochs}, train_accuracy {train_accuracy}, loss {train_loss}, test_accuracy {test_accuracy}, loss {test_loss}, time {time.time() - start}')
# torch.save(pretrained_net, data_name[data_index]+'_resnet50_70epochs.pth')
del pretrained_net1
del pretrained_net2
del lr_scheduler1
del lr_scheduler2
gc.collect()
max_test_acc = max(test_accuracy_list)
print('最终精度：', max_test_acc)


f.close()